from typing import Any
from sp110e.driver import Driver, discover as driver_discover


class Controller:
    """High-level sp110e asynchronous controller.

    Use it only in asynchronous environment (with `asyncio`).

    Author: Pavel Roslovets
    https://roslovets.github.io
    """

    def __init__(self, mac: str, timeout: float = 3.0, retries: int = 0):
        """On object creation."""
        self._mac = mac
        self._timeout = timeout
        self._retries = retries
        self._driver = Driver()

    @staticmethod
    async def discover() -> list:
        """Discover BLE devices."""
        return await driver_discover()

    async def connect(self):
        """Establish BLE connection to device."""
        await self._connect_with_retries()

    async def disconnect(self) -> None:
        """Close connection to device."""
        if self._driver.is_connected():
            await self._driver.disconnect()

    def is_connected(self):
        """Check device is connected."""
        return self._driver.is_connected()

    def is_rgbw(self) -> bool:
        """Check IC model supports RGBW mode."""
        return self._driver.is_ic_model_rgbw()

    async def switch_on(self) -> None:
        """Switch device on."""
        await self.set_state(True)

    async def switch_off(self) -> None:
        """Switch device off."""
        await self.set_state(False)

    async def toggle(self) -> bool:
        """Toggle device state between on/off."""
        new_state = not self.get_state()
        await self.set_state(new_state)
        return new_state

    async def update(self) -> dict:
        """Read device parameters."""
        await self._connect_with_retries()
        return await self._driver.read_parameters()

    async def set_ic_model(self, ic_model: str, force: bool = False) -> None:
        """Set device IC model."""
        await self._connect_and_write_parameter('ic_model', ic_model, force=force)

    async def set_sequence(self, sequence: str, force: bool = False) -> None:
        """Set device color sequence."""
        await self._connect_and_write_parameter('sequence', sequence, force=force)

    async def set_pixels(self, pixels: int, force: bool = False) -> None:
        """Set number of pixels in LED strip."""
        await self._connect_and_write_parameter('pixels', pixels, force=force)

    async def set_state(self, state: bool, force: bool = False) -> None:
        """Set device state: on/off."""
        await self._connect_and_write_parameter('state', state, force=force)

    async def set_mode(self, mode: int, force: bool = False) -> None:
        """Set work mode (0-121). 0 - auto mode (demo)."""
        await self._connect_and_write_parameter('mode', mode, force=force)

    async def set_speed(self, speed: int, force: bool = False) -> None:
        """Set speed of automatic modes (0-255)."""
        await self._connect_and_write_parameter('speed', speed, force=force)

    async def set_brightness(self, brightness: int, force: bool = False) -> None:
        """Set LED brightness (0-255)."""
        await self._connect_and_write_parameter('brightness', brightness, force=force)

    async def set_color(self, color: [int, int, int], force: bool = False) -> None:
        """Set static color in RGB format (0-255)."""
        await self._connect_and_write_parameter('color', list(color), force=force)

    async def set_white(self, white: int, force: bool = False) -> None:
        """Set brightness of white LED (0-255)."""
        await self._connect_and_write_parameter('white', white, force=force)

    def set_mac_address(self, mac_address: str) -> None:
        """Set device MAC address."""
        self._mac = mac_address

    def get_mac_address(self) -> str:
        """Get device MAC address."""
        return self._mac

    def is_on(self) -> bool:
        """Check device is On."""
        return self.get_state()

    def get_ic_model(self) -> str:
        """Get device IC model."""
        return self._driver.get_parameter('ic_model')

    def get_sequence(self) -> str:
        """Set device color sequence."""
        return self._driver.get_parameter('sequence')

    def get_pixels(self) -> int:
        """Set number of pixels in LED strip."""
        return self._driver.get_parameter('pixels')

    def get_state(self) -> bool:
        """Get device state: on/off."""
        return self._driver.get_parameter('state')

    def get_mode(self) -> int:
        """Get work mode (0-121). 0 - auto mode (demo)."""
        return self._driver.get_parameter('mode')

    def get_speed(self) -> int:
        """Get speed of automatic modes (0-255)."""
        return self._driver.get_parameter('speed')

    def get_brightness(self) -> int:
        """Get LED brightness (0-255)."""
        return self._driver.get_parameter('brightness')

    def get_color(self) -> [int, int, int]:
        """Get static color in RGB format (0-255)."""
        return self._driver.get_parameter('color')

    def get_white(self) -> int:
        """Get brightness of white LED (0-255)."""
        return self._driver.get_parameter('white')

    def get_sequences(self) -> tuple:
        """Get list of supported RGB sequence types."""
        return self._driver.get_sequences()

    def get_ic_models(self) -> tuple:
        """Get list of supported IC models."""
        return self._driver.get_ic_models()

    def get_modes(self) -> tuple:
        """Get list of supported modes."""
        return self._driver.get_modes()

    def print_parameters(self):
        """Print device info."""
        self._driver.print_parameters()

    async def __aenter__(self):
        """Enter context callback."""
        await self.connect()
        return self

    async def __aexit__(self, *excinfo):
        """Exit context callback."""
        await self.disconnect()

    async def _connect_with_retries(self):
        """Establish BLE connection to device with retries."""
        for i in range(0, self._retries + 1):
            try:
                if not self._driver.is_connected():
                    await self._driver.connect(self._mac, timeout=self._timeout)
            except Exception:
                if i == self._retries:
                    raise Exception

    async def _connect_and_write_parameter(self, parameter: str, value: Any, force: bool = False) -> None:
        """Write parameter to device with auto connect."""
        if (value is not None) and (force or value != self._driver.get_parameter(parameter)):
            await self._connect_with_retries()
            await self._driver.write_parameter(parameter, value)



from typing import Any, Union
import asyncio
from bleak import BleakClient, BleakScanner, discover as bleak_discover
from bleak.exc import BleakError


async def discover() -> list:
    """Discover BLE devices."""
    devices = await bleak_discover()
    if devices:
        devices_list = [{d.name: d.address} for d in devices]
    else:
        devices_list = None
    return devices_list


class Driver:
    """Low-level sp110e asynchronous BLE driver based on bleak library.

    Use it only if you know why.

    Author: Pavel Roslovets
    https://roslovets.github.io
    """
    IC_MODELS = (
        'SM16703', 'TM1804', 'UCS1903', 'WS2811', 'WS2801', 'SK6812', 'LPD6803', 'LPD8806', 'APA102', 'APA105',
        'DMX512', 'TM1914', 'TM1913', 'P9813', 'INK1003', 'P943S', 'P9411', 'P9413', 'TX1812', 'TX1813',
        'GS8206', 'GS8208', 'SK9822', 'TM1814', 'SK6812_RGBW', 'P9414', 'PG412')
    SEQUENCES = ('RGB', 'RBG', 'GRB', 'GBR', 'BRG', 'BGR')
    CHARACTERISTIC = '0000ffe1-0000-1000-8000-00805f9b34fb'
    MODES = tuple(range(0, 122))
    _client = None
    _parameters = None
    _flag = None

    def __init__(self):
        """Initialize object."""
        self._handle_parameters(bytearray([0] * 12))

    async def connect(self, mac_address: str, timeout: float = 3.0, auto_read: bool = True) -> Union[dict, None]:
        """Establish BLE connection to device."""
        device = await BleakScanner.find_device_by_address(mac_address, timeout=timeout)
        if not device:
            raise BleakError(f'A device with address {mac_address} could not be found')
        self._client = BleakClient(device)
        await self._client.connect()
        await self._client.start_notify(self.CHARACTERISTIC, self._callback_handler)
        if auto_read:
            return await self.read_parameters()
        else:
            return None

    async def disconnect(self) -> None:
        """Close connection to device."""
        await self._client.disconnect()

    def is_connected(self) -> bool:
        """Check connection to device."""
        if self._client:
            return self._client.is_connected
        else:
            return False

    async def send_command(self, command_byte: hex, data_bytes: [hex, hex, hex] = (0x00, 0x00, 0x00)):
        """Send command with data."""
        if type(data_bytes) not in [list, tuple]:
            data_bytes = (data_bytes, 0, 0)
        data_to_write = tuple(data_bytes) + tuple([command_byte])
        await self._client.write_gatt_char(self.CHARACTERISTIC, bytearray(data_to_write))

    async def read_parameters(self) -> dict:
        """Read parameters information from device."""
        self._flag = asyncio.Event()
        await self.send_command(0x10)
        # Wait for callback with data from device
        await asyncio.wait_for(self._flag.wait(), 5)
        return self._parameters

    async def write_parameter(self, parameter: str, value: Any, auto_read: bool = True) -> Union[dict, None]:
        """Write device parameter to device."""
        if parameter == 'state':
            if value:
                await self.send_command(0xAA)  # Switch on
            else:
                await self.send_command(0xAB)  # Switch off
        elif parameter == 'mode':
            if value not in self.MODES:
                raise ValueError(f'Mode is not supported: {value}')
            if value == 0:
                await self.send_command(0x06)
            else:
                await self.send_command(0x2C, value)
        elif parameter == 'speed':
            await self.send_command(0x03, value)
        elif parameter == 'brightness':
            await self.send_command(0x2A, value)
        elif parameter == 'ic_model':
            if value not in self.IC_MODELS:
                raise ValueError(f'IC Model is not supported: {value}')
            idx = self.IC_MODELS.index(value)
            await self.send_command(0x1C, idx)
        elif parameter == 'sequence':
            if value not in self.SEQUENCES:
                raise ValueError(f'Sequence is not supported: {value}')
            idx = self.SEQUENCES.index(value)
            await self.send_command(0x3C, idx)
        elif parameter == 'pixels':
            pixels_bytes = value.to_bytes(2, byteorder='big')
            await self.send_command(0x2D, [pixels_bytes[0], pixels_bytes[1], 0])
        elif parameter == 'color':
            await self.send_command(0x1E, value)
        elif parameter == 'white':
            await self.send_command(0x69, value)
        if auto_read:
            return await self.read_parameters()
        else:
            return None

    async def write_parameters(self, parameters: dict, auto_read: bool = True) -> Union[dict, None]:
        """Write device parameters to device in batch mode."""
        for parameter, value in parameters.items():
            await self.write_parameter(parameter, value, auto_read=False)
        if auto_read:
            return await self.read_parameters()
        else:
            return None

    def get_parameter(self, parameter: str) -> Any:
        """Get read parameter by name."""
        return self._parameters[parameter]

    def get_parameters(self) -> dict:
        """Get read parameters."""
        return self._parameters

    def get_sequences(self) -> tuple:
        """Get list of supported RGB sequence types."""
        return self.SEQUENCES

    def get_ic_models(self) -> tuple:
        """Get list of supported IC models."""
        return self.IC_MODELS

    def is_ic_model_rgbw(self, ic_model: str = None) -> bool:
        """Check IC model supports RGBW mode."""
        if not ic_model:
            ic_model = self.get_parameter('ic_model')
        idx = self.IC_MODELS.index(ic_model)
        return idx > 22

    def get_modes(self) -> tuple:
        """Get list of supported modes."""
        return self.MODES

    def print_parameters(self):
        """Print device info."""
        for key in self._parameters:
            print(key, ':', self._parameters[key])

    def _callback_handler(self, sender, data: bytearray):
        """Handle callback with data from device."""
        if sender == 12:
            self._handle_parameters(data)
        if self._flag:
            # Let get_info() method to go further
            self._flag.set()

    def _handle_parameters(self, data: bytearray):
        """Handle read parameters."""
        self._parameters = {
            'state': data[0] == 1,
            'mode': data[1],
            'speed': data[2],
            'brightness': data[3],
            'ic_model': self.IC_MODELS[data[4]],
            'sequence': self.SEQUENCES[data[5]],
            'pixels': int.from_bytes(data[6:8], byteorder='big'),
            'color': [data[8], data[9], data[10]],
            'white': data[11]
        }



from syncer import sync
from sp110e.controller import Controller


class ControllerSync(Controller):
    """Synchronous adapter for high-level sp110e asynchronous controller.

    Handy tool to use from Python shell or synchronous (normal) env.
    Inherits all synchronous methods from 'controller' class.

    Author: Pavel Roslovets
    https://roslovets.github.io
    """
    Controller = None

    def __init__(self, mac_address: str = None, timeout: float = 3.0, retries: int = 0):
        """On object creation."""
        super().__init__(mac_address, timeout, retries)

    @sync
    async def discover(self) -> list:
        """Discover BLE devices."""
        return await super().discover()

    @sync
    async def connect(self):
        """Establish BLE connection to device."""
        await super().connect()

    @sync
    async def disconnect(self) -> None:
        """Close connection to device."""
        try:
            await super().disconnect()
        except:
            pass

    @sync
    async def switch_on(self) -> None:
        """Switch device on."""
        await super().switch_on()

    @sync
    async def switch_off(self) -> None:
        """Switch device off."""
        await super().switch_off()

    @sync
    async def toggle(self) -> bool:
        """Toggle device state between on/off."""
        return await super().toggle()

    @sync
    async def update(self) -> dict:
        """Read device parameters."""
        return await super().update()

    @sync
    async def set_ic_model(self, ic_model: str, force: bool = False) -> None:
        """Set device IC model."""
        await super().set_ic_model(ic_model, force)

    @sync
    async def set_sequence(self, sequence: str, force: bool = False) -> None:
        """Set device color sequence."""
        await super().set_sequence(sequence, force)

    @sync
    async def set_pixels(self, pixels: int, force: bool = False) -> None:
        await super().set_pixels(pixels, force)

    @sync
    async def set_mode(self, mode: int, force: bool = False) -> None:
        """Set work mode (0-121). 0 - auto mode (demo)."""
        await super().set_mode(mode, force)

    @sync
    async def set_speed(self, speed: int, force: bool = False) -> None:
        """Set speed of automatic modes (0-255)."""
        await super().set_speed(speed, force)

    @sync
    async def set_brightness(self, brightness: int, force: bool = False) -> None:
        """Set LED brightness (0-255)."""
        await super().set_brightness(brightness, force)

    @sync
    async def set_color(self, color: [int, int, int], force: bool = False) -> None:
        """Set static color in RGB format (0-255)."""
        await super().set_color(color, force)

    @sync
    async def set_white(self, white: int, force: bool = False) -> None:
        """Set brightness of white LED (0-255)."""
        await super().set_white(white, force)

    def __del__(self) -> None:
        """On object destruction."""
        self.disconnect()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sp110e']

package_data = \
{'': ['*']}

install_requires = \
['bleak==0.13.0', 'syncer==1.3.0']

setup_kwargs = {
    'name': 'sp110e',
    'version': '1.1.0',
    'description': 'Control SP110E BLE RGB LED device from computer',
    'long_description': "# SP110E Python Library\n\nControl SP110E BLE RGB LED device from computer\n\n## Install\n\n```bash\npip install sp110e\n```\n\n## Tools\n\n- Controller: High-level SP110E asynchronous controller. Use it only in asynchronous environment (with `asyncio`)\n- ControllerSync: Synchronous adapter for high-level SP110E asynchronous controller. Handy tool to use from Python shell or synchronous (normal) environment\n- Driver: Low-level SP110E asynchronous BLE driver based on bleak library. Use it only if you know why\n\n## Documentation\n\n[Full API Reference](./docs)\n\n## Examples\n\nQuick start:\n\n```python\nfrom sp110e.controller_sync import ControllerSync\n\ndevice = ControllerSync('AF:00:10:01:C8:AF')\ndevice.switch_on()\ndevice.set_color([255, 0, 0])\ndevice.set_brightness(255)\n```\n\n[More examples](examples)\n\n## Development\n\n### Create new release\n\nPush changes to 'main' branch following [Conventional Commits](https://www.conventionalcommits.org/en/v1.0.0/).\n\n### Update documentation\n\n`docs` folder is being updated automatically by GitHub Actions when source files are changed.\n\n## Integrations\n\n- [SP110E Home Assistant Integration](https://github.com/roslovets/SP110E-HASS)\n\n## Useful links\n\n- [SP110E API Reference](https://gist.github.com/mbullington/37957501a07ad065b67d4e8d39bfe012)\n- [Vox](https://github.com/nguyenthuongvo/Vox)\n- [bleak library](https://github.com/hbldh/bleak)\n- [Reverse engineering simple BLE](http://nilhcem.com/iot/reverse-engineering-simple-bluetooth-devices)\n",
    'author': 'Pavel Roslovets',
    'author_email': 'p.v.roslovets@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/roslovets/SP110E',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)
